/* 

Copyright (c) 1999  Ulric Eriksson <ulric@siag.nu>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <magick/magick.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <limits.h>

#ifndef USHRT_MAX
#define USHRT_MAX 65535
#endif

/* known colors are cached in a linked list to avoid
   making multiple calls to the X server */
typedef struct xmagick_color {
	XColor color;
	struct xmagick_color *next;
} xmagick_color;

static xmagick_color *known_colors = NULL;

/* forget all cached colors */
static void free_colors(void)
{
	xmagick_color *p = known_colors;
	while (p) {
		known_colors = p->next;
		free(p);
		p = known_colors;
	}
	known_colors = NULL;
}

/* call XAllocColor through the cache */
/* Does not fail. Returns closest match at worst. */
static void alloc_color(Display *display, Colormap colormap, XColor *color)
{
	xmagick_color *p = known_colors;

	while (p) {
		if (p->color.red == color->red
				&& p->color.green == color->green
				&& p->color.blue == color->blue) {
			color->pixel = p->color.pixel;
			return;
		}
		p = p->next;
	}
	p = malloc(sizeof *p);
	p->color = *color;
	if (XAllocColor(display, colormap, color) == 0) {
		/* failed, must try harder */
		color->pixel = 0;	/* ;-) */
	}
	p->color.pixel = color->pixel;
	p->next = known_colors;
	known_colors = p;
}

/* call XQueryColor through the cache */
static void query_color(Display *display, Colormap colormap, XColor *color)
{
	xmagick_color *p = known_colors;

	while (p) {
		if (p->color.pixel == color->pixel) {
			color->red = p->color.red;
			color->green = p->color.green;
			color->blue = p->color.blue;
			return;
		}
		p = p->next;
	}
	p = malloc(sizeof *p);
	XQueryColor(display, colormap, color);
	p->color = *color;
	p->next = known_colors;
	known_colors = p;
}

/* Convert Image to XImage */
XImage *XMagickImageToXImage(Display *display, Image *image)
{
	XColor color;
	unsigned long pixel;
	int i;
	float *red, *green, *blue, *opacity;
	int x, y;
	XImage *ximage = NULL;
	int screen = DefaultScreen(display);
	Visual *visual = DefaultVisual(display, screen);
	int format = ZPixmap;
	int offset = 0;
	char *data = NULL;
	Colormap colormap = DefaultColormap(display, screen);
	int bitmap_pad = XBitmapPad(display);
	int bytes_per_line = 0;
	unsigned int width = image->columns;
	unsigned int height = image->rows;
	unsigned int depth = DefaultDepth(display, screen);

	free_colors();
	red = malloc((width*height)*(sizeof *red));
	green = malloc((width*height)*(sizeof *green));
	blue = malloc((width*height)*(sizeof *blue));
	opacity = malloc((width*height)*(sizeof *opacity));

	if (red && green && blue && opacity) {
		GetPixels(image, red, green, blue, opacity);
		ximage = XCreateImage(display, visual, depth,
				format, offset, data,
				width, height, bitmap_pad, bytes_per_line);
		ximage->data = malloc(ximage->bytes_per_line
					*ximage->height*ximage->depth);
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				i = x+y*width;
				color.red = USHRT_MAX*red[i];
				color.green = USHRT_MAX*green[i];
				color.blue = USHRT_MAX*blue[i];
				alloc_color(display, colormap, &color);
				pixel = color.pixel;
				XPutPixel(ximage, x, y, pixel);
			}
		}
	}

	free(red);
	free(green);
	free(blue);
	free(opacity);

	free_colors();
	return ximage;
}

/* Convert XImage to Image */
Image *XMagickXImageToImage(Display *display, XImage *ximage)
{
	XColor color;
	unsigned long pixel;
	int i;
	float *red, *green, *blue, *opacity;
	Image *image = NULL;
	int x, y;
	int screen = DefaultScreen(display);
	Colormap colormap = DefaultColormap(display, screen);
	unsigned int width = ximage->width;
	unsigned int height = ximage->height;

	free_colors();
	red = malloc((width*height)*(sizeof *red));
	green = malloc((width*height)*(sizeof *green));
	blue = malloc((width*height)*(sizeof *blue));
	opacity = malloc((width*height)*(sizeof *opacity));

	if (red && green && blue && opacity) {
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				i = x+y*width;
				pixel = XGetPixel(ximage, x, y);
				color.pixel = pixel;
				query_color(display, colormap, &color);
				red[i] = (float)color.red/USHRT_MAX;
				green[i] = (float)color.green/USHRT_MAX;
				blue[i] = (float)color.blue/USHRT_MAX;
				opacity[i] = 1.0;
			}
		}
		image = CreateImage(width, height, red, green, blue, opacity);
	}

	free(red);
	free(green);
	free(blue);
	free(opacity);

	free_colors();
	return image;
}

